/*
   $Id: ParametersChain.cpp 666 2011-11-15 10:16:53Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#include <string>
#include <list>
#include <iostream>
#include <vector>
using namespace std;

#include "ptypes.h"
#include "SimpleOpt.h"

#include "ParametersChain.h"
#include "CommandChain.h"

CSimpleOpt ParametersInitializerChain::_Init(int argc, char* argv[], const Command& cmd)
{
	CSimpleOpt::SOption options[] = {
	    { OPT_RESULTSFILENAME,  (char*)"-o", SO_REQ_SEP }, // "-o"
	    { OPT_RESULTSFILENAME,  (char*)"--output-file", SO_REQ_SEP }, // "--output-file"
	    { OPT_SELF,  (char*)"-S",         SO_NONE    }, // "-S"
	    { OPT_SELF,  (char*)"--self",     SO_NONE    }, // "--self"
	    { OPT_SILENT,  (char*)"--silent",   SO_NONE }, //
	    { OPT_VERBOSE,  (char*)"--verbose", SO_NONE }, //
	    { OPT_VERBOSE,  (char*)"-v",        SO_NONE }, //
	
		// writing compressed data
		{ OPT_COMPRESS,   (char*) "--compress",     SO_REQ_SEP },
		
	    //For the chaining algorithm
	    { OPT_CUTOFFS,    (char*)"--cutoffs",     SO_REQ_SEP    },
	    { OPT_MAXOVERLAP, (char*)"--maxoverlap",     SO_REQ_SEP    },
	    { OPT_GLINTOUT,   (char*)"--glintout",     SO_REQ_SEP    },
	    { OPT_BLASTPOUT,  (char*)"--blastpout",     SO_REQ_SEP    },
	    { OPT_ANCHORSOUT, (char*)"--anchorsout",     SO_REQ_SEP    },
	    { OPT_TARGETDIR,  (char*)"--targetdir",     SO_REQ_SEP    },
	    { OPT_NORBH,      (char*)"--noRBH",     SO_NONE    }, //
	    
	    // To get some help
	    { OPT_HELP, (char*)"-?",     SO_NONE    }, // "-?"
	    { OPT_HELP, (char*)"-h",     SO_NONE    }, // "-?"
	    { OPT_HELP, (char*)"--help", SO_NONE    }, // "--help"
	    
	    SO_END_OF_OPTIONS                       // END
	};

	//p.__Help = help;
	p.__argc = argc;

	// declare our options parser, pass in the arguments from main
    // as well as our array of valid options.
    CSimpleOpt args(argc, argv, options);
	while (args.Next()) {
		if (args.LastError() == SO_SUCCESS) 
		{
			if (args.OptionId() == OPT_HELP) 
			{
				cmd.Usage();
				exit(0);
			}
			else if ( args.OptionId() == OPT_GLINTOUT )
			{
				p.__glint_output_filename = (CompressedFileName) args.OptionArg();
			}
			else if ( args.OptionId() == OPT_BLASTPOUT )
			{
				p.__blastp_output_filename = (CompressedFileName) args.OptionArg();
			}
			else if ( args.OptionId() == OPT_ANCHORSOUT )
			{
				p.__anchors_output_filename = (CompressedFileName) args.OptionArg();
			}
			else if (args.OptionId() == OPT_CUTOFFS )
			{
				p.__chain_cutoffs = (string) args.OptionArg();
			}
			else if (args.OptionId() == OPT_MAXOVERLAP )
			{
				p.__max_overlap = atoi(args.OptionArg());
			}
			else if (args.OptionId() == OPT_TARGETDIR )
			{
				p.__target_dir = (string) args.OptionArg();
			}
			else if (args.OptionId() == OPT_NORBH )
			{
				p.__rbh = false;
			}
			else if (args.OptionId() == OPT_SELF )
			{
				p.__self = true;
			}
			else if (args.OptionId() == OPT_RESULTSFILENAME )
			{
				p.__results_filename = (CompressedFileName) args.OptionArg();
			}
			else if (args.OptionId() ==  OPT_SILENT )
          	{
            	p.SetSilent();
          	}
			else if ( args.OptionId() == OPT_COMPRESS )
			{
				p.__results_filename.SetCompression(args.OptionArg());
			}
			#ifdef DEBUGLOG
			else if (args.OptionId() ==  OPT_LOG_LEVEL)
			{
				p.__log_level = (string) args.OptionArg();
			}
			#endif
		}
        else 
        {
            fprintf(stderr,"ERROR - Invalid argument: %s\n", args.OptionText());
            fprintf(stderr,"Please use option --help to see the usage\n");
            exit(1);	
        }
    }
    return args;
}

/**
 * \brief called by _Init to finish the job
 */
void ParametersInitializerChain::_Init2 ( const CSimpleOpt& args, const Command& cmd ) {}

/** 
 *  \brief Validate the parameters useful for glintindex
 *  \return true if all good
*/
bool ParametersInitializerChain::_Validation()
{
	bool ok = true;
	if ( p.__glint_output_filename.GetName() == "" && p.__blastp_output_filename.GetName() == "" && p.__anchors_output_filename.GetName() == "" )
	{
		cerr << "ERROR - At least one of glint out, blast out, or anchors filename must be provided.\n";
		ok = false;
	}

    if ( p.__anchors_output_filename.GetName() != "" && ( p.__glint_output_filename.GetName() != "" || p.__blastp_output_filename.GetName() != "" ) )
    {
    	cerr << "ERROR - You must choose between anchors or glint and blastp.\n";
    	ok = false;
    }

    if ( p.__chain_cutoffs == DEFAULT_CHAINING_CUTOFFS )
    {
		if (!p.IsSilent())
		{
			cerr << "WARNING - Using default cutoff parameters (" << DEFAULT_CHAINING_CUTOFFS << ").\n";
		}
    }

    if ( ! __ValidCutoffParam( p.__chain_cutoffs ) )
    {
    	cerr << "ERROR - Something wrong with the cutoff parameters.\n";
    	ok = false;
    }
	return ok;
}

/**
 * \brief Validate the format or the cutoffs parameters
 * \param cutoff: the cutoffs, formatted as explained here
 */
/*
 cutoffs : a string that represents the cutoffs formated as explained below (see also the help)
           The cutoff values are given for each level as 3 values using the following format CMD-CS-CCV :
               CMD : cutoff value for the manhattan Distance separating anchors (integer, a strict maximum value)
                     (will be used for the construction of connected components)
               CS  : cutoff score for a cluster to be part of the cluster partition (integer, a strict minimum value)
               CCV : cutoff score for the covariation of anchors within a cluster (float, a strict minimum value)
 Return		:  false if one of the cutoff parameters are not correctly read
*/

bool ParametersInitializerChain::__ValidCutoffParam( const string cutoffs )
{
	bool ok = true;

	list<string> cutoffs_list;

	std::istringstream iss( cutoffs );
	std::string level_token;
	while ( getline(iss,level_token, ':') )
	{
		cutoffs_list.push_back( level_token );
	}

	for( list<string>::iterator i = cutoffs_list.begin() ; i!=cutoffs_list.end() ; ++i )
	{
		std::istringstream iss_cutoff( *i );
		std::string cutoff_token;
		vector<string> cutoff_vector;
		while ( getline(iss_cutoff,cutoff_token, '-') )
		{
			cutoff_vector.push_back( cutoff_token );
		}
		if ( cutoff_vector.size() != 3 )
		{
			ok = false;
			break;
		}

		// Checks if all the cutoff parameters are OK
		ok &= Validuint32_t(cutoff_vector[0]) && Validuint32_t(cutoff_vector[1]) && ValidCovar(cutoff_vector[2]);
	}
	return ok;
}
