/*
   $Id: QueryBuffer.h 649 2011-10-31 13:57:42Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef _QUERYBUFFER_H_
#define _QUERYBUFFER_H_

#include <list>
using namespace std;

#include "../samtools/sam.h"
#include "../samtools/bam.h"
#include "rtstack.h"
#include "Parameters.h"
#include "buffer.h"
#include "hsp.h"
#include "SeqInfo.h"


/** \brief the QueryBuffer hierarchy
 *         - These classes define a buffer, to be filled with characters from the query sequence.
 *         - QueryBuffer              The base, abstract class
             - QueryBufferFile        Reads the query sequence from a multifasta fasta file
               -    QueryBufferBank   Reads the query sequence from a .fa file, the sequence id from a .he file
                                      both produced by glintindex (useful with the switch -G)
             - QueryBufferSelf        Reads the query sequence from a bank already in memory (useful with the switch -S)

           - Public methods:
             - ReadNextSeq()   Load the buffer with the next sequence
             - GetSeqLength    Returns the length of the loaded sequence
             - GetSeqId()      Returns the loaded sequence id
             - GetSeq()        Returns a char* representing the loaded sequence
             - GetseqAsFasta() Same as GetSeqId()/GetSeq, fasta-formatted
             - Getc(int i)     Returns the ith character of the sequence, 1st base for i = 1

*/

/** \brief The base class
 */

class QueryBuffer {
	public:

		QueryBuffer(): 
		_parameters(Parameters::Instance()),
		__sequence_id(""), 
		__sequence_length(0), 
		__sequence_clipped(0), 
		__fragnum(0) {};
		virtual ~QueryBuffer(){};
		virtual QueryBuffer* Clone() = 0;
			
		// Read the next seq - 
		// Return true if a new sequence is read, and update fields to that the accessors return valid data
		// Return false if there is no more sequence to read, the accessors return the LAST read sequence
	    virtual bool ReadNextSeq()=0;
	
	    /* Several accessors */
		size_t GetSeqLength() const          { return __sequence_length; };
		
		// NOTE - If the sequence is clipped, GetSeq returns a fake address (BEFORE the buffer base - Use at your own risk !)
	    // const unsigned char* GetSeq() const { return __buffer_base - __sequence_clipped;   };

		// NOTE - You should use this function whenever possible
		// NOTE - First base => i = 1, relative to the sequence
		// TODO - We SHOULD check the overflow here
		unsigned char Getc(uint32_t i) const    {return *(__buffer_base - __sequence_clipped + i);}; 
	    
	    // NOTE - This is much better, as it points to a REAL memory area !
	    const unsigned char* GetBase() const { return __buffer_base; };
	    
	    // Only several input formats handle quality data - the default is to return NULL
	    virtual const char * GetQual()      const { return NULL; };
	    
	    // Only SAM/BAM input format handles auxiliary data - the default is to return NULL
	    virtual const uint8_t* GetAuxData()       const { return NULL; };
		virtual int            GetAuxDataLength() const { return 0; };
		
		// Only SAM/BAM input format handles the subject - the default is to return NULL
		virtual const unsigned char* GetSubjectBase() const { return NULL; };
		
		// Only several input formats contain a real hsp - the default is to return NULL
		virtual const Hsp* GetHsp() const { return NULL; };
		
		// Used only by sam/bam input formats
		virtual void ReverseCompl() {
			ReverseComplement& reverse_complement = ReverseComplement::Instance();
			reverse_complement(__buffer_base+1,__sequence_length);
		};
		
		auto_ptr<string> GetSeqAsFasta() const ;
		auto_ptr<string> GetSeqOneLine() const ;
		string GetSeqId() const              { return __sequence_id; };
		
		/* call this function to make the letters be converted to upper case */
		static void ToUpper(bool to_up)  { _h_flg = (signed char) to_up ? 0xDF : 0xFF; };

		// Maximum possible score of the query sequence given its length.
		// matchscore is the atomic reward for a nucleotide match
	    uint64_t MaxScore(int matchscore) const { return ( GetSeqLength()*matchscore ); }; 

		// Dummy functions, see the implementation in QueryBufferFile
		virtual bool IsLower(uint8_t ignored=0)   const { return true; };
		virtual bool IsUpper(uint8_t ignored=0)   const { return true; };
		
		//Setting and Getting the fragment number of the template (first or second <=> fwd and bwd)
		void SetFragNum(uint8_t fragnum) {  __fragnum = fragnum; };
		uint8_t GetFragNum()   const     {  return __fragnum; };
		
		//Temporary Patch enabling to rename a sequence (adding fwd and bwd when reading from the bam format)
		void ChangeSeqId(const string & n) {  __sequence_id     = n; };
			
	protected:

	    // Setting the sequence id, length, base address. This is used by the derived classes
	    void _SetSeqId(const string & n) { __sequence_id     = n; };
	    void _SetSeqLength(size_t l)     { __sequence_length = l; };
	    void _SetBase(unsigned char *c)  { __buffer_base     = c; };
	    void _SetClip(size_t l)          { __sequence_clipped= l; };
		static signed char _h_flg;       /* may be $FF (do not change) or $DF (change to upper) */
		Parameters& _parameters;
		
    private:
	    //Parameters& __parameters;
		string __sequence_id;
		size_t __sequence_length;
		size_t __sequence_clipped;
		unsigned char * __buffer_base;
		
		// When sequences where produced in paired-end or mate-pair
		uint8_t  __fragnum;	
};

/**
 * \brief QueryBufferFileHandler: 
 *             we maintain a static vector of file handlers (may be used as FILE*, as samfile_t* etc.)
 *             The constructor pushes a new element in this vector, the derived classes open the file using the last pushed pointer
 *             The COPY constructor (generally used through the derived Clone() functions) ONLY COPY the resulting whatever* file descriptor
 *             So, several cloned QuerybufferFile will read THE SAME file.
 *             This is used by the QBufferManager found in Comparison.cpp
 *             If you construct several QueryBufferFiles with several file names, each file 
 *             shall have a different file descriptor. This is used in QueryBufferdouble
 * 
 *             The Clone() method implements a polymorphic clone pattern
 */

class QueryBufferFileHandler: public QueryBuffer
{
	public:
		QueryBufferFileHandler();
		virtual ~QueryBufferFileHandler(){};
		
	protected:
		/**
		 * \brief Several types of file descriptors are hidden inside this union
		 */		
		struct FileDescPtr
		{
			FileDescPtr():fd(NULL),sam_fd(NULL){};
			~FileDescPtr(){FClose(fd);if(sam_fd){samclose(sam_fd);};};
			FILE* fd;
			samfile_t* sam_fd;
		};

	    static list<FileDescPtr> _file_descriptors;
	    
};

/**
 * \brief QueryBufferFile: 
 *             Used to fill the QueryBuffer from a file, extending QueryBufferfileHandler
 */

class QueryBufferFile: public QueryBufferFileHandler
{
	public:
		QueryBufferFile();
		virtual ~QueryBufferFile(){};
		
		// Return true if there is at least one lowercase in the sequence
		bool IsLower(uint8_t ignored=0)   const { for (size_t i=1; i<=GetSeqLength()-ignored; ++i) {if ((Getc(i)&0x20) != 0)return true;};return false;};
		// Return true if there is at least one uppercase in the sequence
		bool IsUpper(uint8_t ignored=0)   const { for (size_t i=1; i<=GetSeqLength()-ignored; ++i) {if ((Getc(i)&0x20) == 0)return true;};return false;};
	
	protected:
	    /* Putting a new character in the buffer - The buffer is resized if necessary */
	    void _Putc(char c) {
		    if (c!='\n' && (signed char) c!= EOF)
		    {
				c &= _h_flg;
		        _query_bfr.Put(c);
			}
		};
		void _PutSeparator() {_query_bfr.Put('$');};

	    /* Clearing the buffer for a new sequence, putting a $ at start */
	    void _Clear() {
		    _query_bfr.Clear();
			_PutSeparator();
		};
	
	   Buffer _query_bfr;
   
};

/**
 * \brief QueryBufferFasta: used to fill the QueryBuffer from a fasta file
 */

class QueryBufferFasta: public QueryBufferFile
{
	public:
		// bool qual=false is a dummy parameter which will always be false (for compatibility reason with QueryBufferFastq
		QueryBufferFasta(const string&, CompressionAlgorithm = NOCOMPRESSION, bool qual=false);
		~QueryBufferFasta()  { FClose( __fd_in ); };
		virtual QueryBufferFasta* Clone() { return new QueryBufferFasta( *this ); };
		
		virtual bool ReadNextSeq();
			
	private:
		FILE*& __fd_in;          // The input file descriptor

};

/** 
 * \brief QueryBufferQual: a base class for QueryBufferFile with quality (fastq, sam/bam)
 *                          quality_flg should be set to true if quality must be kept in the object (not always useful)
 */
class QueryBufferQual: public QueryBufferFile
{
	public:
		QueryBufferQual(bool quality_flg=false):
			_quality_flg(quality_flg) {};
	    virtual const char * GetQual() const 
	    {
			//const unsigned char* qual = NULL;
			return _quality_flg ? (const char*) _quality_bfr.GetBase(): NULL;
		};
	
	protected:
		Buffer _quality_bfr;	// Optionaly keeping the information about quality (not used but may be useful for output)
		bool   _quality_flg;
};

/**
 * \brief QueryBufferSam: used to fill the QueryBuffer from a Sam/Bam file
 */
class QueryBufferSam: public QueryBufferQual
{
	public:
		QueryBufferSam(const string&, bool bam_flg=false, bool qual=false, bool hsp=false);
		QueryBufferSam(bam_header_t*, bool bam_flg=false, bool qual=false, bool hsp=false);
		~QueryBufferSam();
		virtual QueryBufferSam* Clone() { return new QueryBufferSam( *this ); };
		virtual const Hsp* GetHsp() const { return __hsp_aptr.get(); };
		Hsp* GetHsp() { return __hsp_aptr.get(); };  // version non const - used by QueryBufferSamDouble
	    virtual const uint8_t* GetAuxData()           const { return __aux_data; };
		virtual int            GetAuxDataLength()     const { return __aux_data_len;   };
		virtual const unsigned char* GetSubjectBase() const { return (const unsigned char*) __subject_bfr.GetBase(); };
		uint16_t GetFlag() const { return __flag; };
		
		virtual bool ReadNextSeq();
		void ReadFromLowLevelBuffer(const bam1_t* b);
		
	private:
		samfile_t*& __fd_in;        // The input file descriptor
		bam_header_t *__header;
		bool __bam_flg;
		long __file_pos;
		bool __hsp_flg;
		auto_ptr<Hsp> __hsp_aptr;
		uint8_t*      __aux_data;
		int  __aux_data_len;
		
		void __InitHsp       ( const bam1_t* );
		void __RevertCigar   ( const bam1_t* );
		void __ReadQual      ( const bam1_t* );
		void __ReadAux       ( const bam1_t* );
		void __ReadSubject   ( const bam1_t* b);
		string __RevertMD    ( const string& ) const;
		Buffer __subject_bfr;
		uint16_t __flag;
};

/**
 * \brief QueryBufferBamLowLevel: A LowLevel QueryBufferBam, only TWO methods are implemented:
 *        -readNextSeq()
 *        -GetLowLevelBuffer()
 */
class QueryBufferBamLowLevel: public QueryBufferFile
{
	public:
		QueryBufferBamLowLevel(const string&);
		~QueryBufferBamLowLevel();
		virtual QueryBufferBamLowLevel* Clone() { return new QueryBufferBamLowLevel( *this ); };
		const bam1_t* GetLowLevelBuffer() const { return __b; };
		virtual bool ReadNextSeq();
	    bam_header_t* GetHeader() const { return __fd_in -> header; };

	private:
		samfile_t*& __fd_in;        // The input file descriptor
		bam1_t* __b;
};

/**
 * \brief QueryBufferFastq: used to fill the QueryBuffer from a fastq file
 */

class QueryBufferFastq: public QueryBufferQual
{
	public:
		QueryBufferFastq(const string&, CompressionAlgorithm = NOCOMPRESSION, bool qual=false);
		~QueryBufferFastq()  { FClose( __fd_in ); };
		virtual QueryBufferFastq* Clone() { return new QueryBufferFastq( *this ); };
		
		virtual bool ReadNextSeq();
			
	private:
		FILE*& __fd_in;          // The input file descriptor

};

#endif
