/*
   $Id: QueryBufferBank.h 649 2011-10-31 13:57:42Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef _QUERYBUFFERBANK_H_
#define _QUERYBUFFERBANK_H_

#include <list>
using namespace std;
#include "QueryBuffer.h"


/** \brief the QueryBuffer hierarchy
 *         - These classes define a buffer, to be filled with characters from the query sequence.
 *         - QueryBuffer              The base, abstract class
             - QueryBufferFile        Reads the query sequence from a multifasta fasta file
               -    QueryBufferBank   Reads the query sequence from a .fa file, the sequence id from a .he file
                                      both produced by glintindex (useful with the switch -G)
             - QueryBufferSelf        Reads the query sequence from a bank already in memory (useful with the switch -S)

           - Public methods:
             - ReadNextSeq()   Load the buffer with the next sequence
             - GetSeqLength    Returns the length of the loaded sequence
             - GetSeqId()      Returns the loaded sequence id
             - GetSeq()        Returns a char* representing the loaded sequence
             - GetseqAsFasta() Same as GetSeqId()/GetSeq, fasta-formatted
             - Getc(int i)     Returns the ith character of the sequence, 1st base for i = 1

*/


/**
 * \brief QueryBufferBank: used to fill the QueryBuffer from a .fa file (-G switch)
 */

class QueryBufferBank: public QueryBufferFile
{
	public:
		QueryBufferBank(const string&, CompressionAlgorithm = NOCOMPRESSION );
		virtual ~QueryBufferBank() { FClose( __fd_in ); };
		virtual QueryBufferBank* Clone() { return new QueryBufferBank( *this ); };
		virtual bool ReadNextSeq();
	
	private:
		SeqInfo seq_info;
		uint64_t seq_cnt;
		uint64_t seq_nb;
		FILE*& __fd_in;          // The input file descriptor

};

/**
 * \brief QueryBufferSelf: used to fill the QueryBuffer from a bank already in memory
 *                         Extends QueryBuffer, this is useful with the -S switch (self comparison)
 */

class CompactTree;
class QueryBufferSelf: public QueryBuffer
{
	public:
		QueryBufferSelf();
		virtual ~QueryBufferSelf() {};
		virtual bool ReadNextSeq();
		virtual QueryBufferSelf* Clone() 
		{ 
			throw ( logic_error( (string) "ERROR - QueryBufferSelf cannot be cloned !" + ERROR_FILE_LINE(__FILE__,__LINE__)));
		};
	
	private:
		const CompactTree& __ctree;
};

#endif
