/* $Id: QueryBufferDouble.cpp 697 2013-01-22 11:48:29Z manu $ */
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#include <vector>
#include <string>
#include <iostream>
#include <iomanip>
#include "utils.h"
#include "../samtools/sam.h"
#include "QueryBufferDouble.h"

using namespace std;

extern RealLogger* logger;

QueryBufferSamDouble::QueryBufferSamDouble(const string& filename, bool qual, bool hsp):
__low_level(filename),
__fwd(__low_level.GetHeader(),true,qual,hsp),
__bwd(__low_level.GetHeader(),true,qual,hsp),
__hsp_flg(hsp)
{__fwd.SetFragNum(FWD);__bwd.SetFragNum(BWD);}
map<string,bam1_t*> QueryBufferSamDouble::__orphans;


/**
 * \brief Read the next sequence from the bam file, if it is the mate of an already read sequence, return true
 *        Else, keep the sequence in memory, waiting for its mate, and read the next one
 * \return true if pair sequence was read, false if not (End of file)
 */
bool QueryBufferSamDouble::ReadNextSeq()
{
	bool found_a_mate = false;
	bool ok = false;
			
	do
	{
		ok = __low_level.ReadNextSeq();
		if ( ok ) 
		{
			bam1_t* b = bam_dup1(__low_level.GetLowLevelBuffer());
			
			string id = (const char*)bam1_qname(b);
			
			map<string,bam1_t*>::iterator i=__orphans.find(id);
			if ( i != __orphans.end() )
			{	
				// We are reading 'new' data from a pair, other data is already 'found'
				// But who is fwd ? who is bwd ? The flag should help
				uint32_t flag_found  = (i->second->core.flag & (BAM_FREAD1 + BAM_FREAD2))>>6;
				uint32_t flag_new    = (b->core.flag         & (BAM_FREAD1 + BAM_FREAD2))>>6;
				uint32_t flag        = (flag_found<<2) + flag_new;
				if ( i->second->core.flag&BAM_FPAIRED && b->core.flag&BAM_FPAIRED )
				{
					// flag = 0110 ==> found is fwd, new is bwd
					if ( flag == 6 )
					{
						__fwd.ReadFromLowLevelBuffer(i->second);
						__bwd.ReadFromLowLevelBuffer(b);
					}
					
					// flag = 1001 ==> found is bwd, new is fwd
					else if ( flag == 9 )
					{
						__bwd.ReadFromLowLevelBuffer(i->second);
						__fwd.ReadFromLowLevelBuffer(b);
					}
					
					// flag = 0000 ==> both are nothing - this could happen, but we keep track of it
					//                 We choose ourself: let's say that found is fwd, new if bwd
					else if ( flag == 0 )
					{
						__fwd.ReadFromLowLevelBuffer(i->second);
						__bwd.ReadFromLowLevelBuffer(b);
						//cerr << "WARNING - Do not know which pair is fwd and which is bwd " << id << "\n";
						_parameters.Warn("Do not know which pair is fwd and which is bwd",id);
					}
					
					// flag = any other value ==> THIS SHOULD NOT HAPPEN - KEEP TRACK OF IT !
					else
					{
						__fwd.ReadFromLowLevelBuffer(i->second);
						__bwd.ReadFromLowLevelBuffer(b);
						//cerr << "WARNING - BAM FLAG INCOHERENT (both pairs are fwd AND bwd !) " << id << "\n";
						_parameters.Warn("BAM FLAG INCOHERENT (both pairs are fwd AND bwd !)",id);
					}
				}
				
				// At least one of them is not in a pair - this SHOULD NOT HAPPEN !!!
				else
				{
					__fwd.ReadFromLowLevelBuffer(i->second);
					__bwd.ReadFromLowLevelBuffer(b);
					//cerr << "WARNING - BAM FLAG INCOHERENT (read does not belong to a pair) " << id << "\n";
					_parameters.Warn("BAM FLAG INCOHERENT (read does not belong to a pair)",id);
				}									
				 			
				// Init the querybuffers from the low level bam buffer
				//__fwd.SetFragNum(FWD);
				//__bwd.SetFragNum(BWD);
				
				//__bwd.ReadFromLowLevelBuffer(i->second);
				//__fwd.ReadFromLowLevelBuffer(b);
				
				_SetSeqId     (i->first);
				_SetBase      ((unsigned char*)__fwd.GetBase());
				_SetSeqLength (__fwd.GetSeqLength());
				found_a_mate = true;
				
				// Destroy the buffers and remove the sequence from __orphans, as we found its mate !
				bam_destroy1(i->second);
				bam_destroy1(b);
				__orphans.erase(i);
				//cerr << setw(40) << "*** FOUND A MATE " << setw(50) << id << " NB OF ORPHANS: " << __orphans.size() << "\n";
	
				// Declare that the hsp are paired
				if ( __hsp_flg )
				{
					Pair( __fwd.GetHsp(), __bwd.GetHsp() );
				}

			}
			else
			{
				__orphans[id] = b;
#ifdef DEBUGLOG	
				cerr.setf(ios::left,ios::adjustfield);
				cerr << setw(40) << "*** FOUND AN ORPHAN " << setw(50) << id << " NB OF ORPHANS NOW: " << __orphans.size() << "\n";
#endif
			}
		}
	}
	while ( ok && !found_a_mate );
	return ok;
}


/**
 * \brief Constructor
 */

QueryBufferBamLowLevelDouble::QueryBufferBamLowLevelDouble(const string& filename): 
__fwd(NULL),
__bwd(NULL),
__low_level(filename) {}
map<string,bam1_t*> QueryBufferBamLowLevelDouble::__orphans;


/**
 * \brief Read the next sequence from the bam file, if it is the mate of an already read sequence, return true
 *        Else, keep the sequence in memory, waiting for its mate, and read the next one
 * \return true if pair sequence was read, false if not (End of file)
 */
bool QueryBufferBamLowLevelDouble::ReadNextSeq()
{
	bool found_a_mate = false;
	bool ok = false;
	do
	{
		ok = __low_level.ReadNextSeq();
		if ( ok ) 
		{
			bam1_t* b = bam_dup1(__low_level.GetLowLevelBuffer());
			
			string id = (const char*)bam1_qname(b);
			
			map<string,bam1_t*>::iterator i=__orphans.find(id);
			if ( i != __orphans.end() )
			{
				if ( __fwd != NULL )
				{
					bam_destroy1(__fwd);
				}
				if ( __bwd != NULL )
				{
					bam_destroy1(__bwd);
				}
				__fwd = i->second;
				__bwd = b;
				_SetSeqId     (i->first);
				found_a_mate = true;
				
				// remove the sequence form __orphans, as we found its mate !
				__orphans.erase(i);
#ifdef DEBUGLOG	
				cerr.setf(ios::left,ios::adjustfield);
				cerr << setw(40) << "*** FOUND A MATE " << setw(50) << id << " NB OF ORPHANS NOW: " << __orphans.size() << "\n";
#endif	
			}
			else
			{
				__orphans[id] = b;
#ifdef DEBUGLOG	
				cerr.setf(ios::left,ios::adjustfield);
				cerr << setw(40) << "*** FOUND AN ORPHAN " << setw(50) << id << " NB OF ORPHANS NOW: " << __orphans.size() << "\n";
#endif
			}
		}
	}
	while ( ok && !found_a_mate );
	return ok;
}

