/*
   $Id: QueryBufferDouble.h$
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/

#ifndef _QUERYBUFFERDOUBLE_H_
#define _QUERYBUFFERDOUBLE_H_

#include <list>
using namespace std;

#include "../samtools/sam.h"
#include "rtstack.h"
#include "Parameters.h"
#include "buffer.h"
#include "hsp.h"
#include "SeqInfo.h"
#include "utils.h"
#include "QueryBuffer.h"

/**
 * \brief QueryBufferDouble: This QueryBuffer is useful for managing pair-ended sequences
 *                           In contains two identical queryBuffer-derived objects
 *                           Both files should have the same format and the same compression level
 *                           GetSeqId() returns the common Id
 *                           All other accessors have their GetxxxBwd() counterpart
 *                           ReadNextSeq() calls both ReadNextSeq
 *                           QueryBufferDouble is the interface, common to ALL QueryBufferXXXDouble classes
 */

class QueryBufferDouble: public QueryBufferFileHandler
{
	public:
		QueryBufferDouble() : _fd_in(_file_descriptors.back().fd) {};
		virtual ~QueryBufferDouble(){};
		virtual size_t GetSeqLengthBwd() const            = 0;
	    virtual const unsigned char * GetBaseBwd() const  = 0;
		virtual auto_ptr<string> GetSeqAsFastaBwd() const = 0;
		virtual unsigned char GetcBwd(uint32_t i) const      = 0;
	    virtual uint64_t MaxScoreBwd(int matchscore) const   = 0;
		virtual const Hsp* GetHspMate() const { return NULL; };   // See GetHsp() from QueryBuffer	
	    virtual QueryBuffer& GetFwdQueryBuffer()  = 0;
	    virtual QueryBuffer& GetBwdQueryBuffer() = 0;	
	    //virtual bool ValidPairEndIds(const string& fwdId, const string& bwdId );
	    
	protected:
		FILE*& _fd_in;
};

template < typename T > class QueryBufferFormatDouble: public QueryBufferDouble
{
	public:
		QueryBufferFormatDouble (const string& filename, const string& filename_mate, CompressionAlgorithm compression=NOCOMPRESSION, bool qual=false):
		__fwd(filename, compression, qual), __bwd(filename_mate, compression, qual) { };
		virtual ~QueryBufferFormatDouble(){};
		virtual QueryBufferFormatDouble* Clone() { return new QueryBufferFormatDouble<T>( *this ); };
		T& GetFwdQueryBuffer() { return __fwd; };
		T& GetBwdQueryBuffer() { return __bwd; };
					
	    bool ReadNextSeq()
	    {
			bool read      = __fwd.ReadNextSeq();
			bool read_mate = __bwd.ReadNextSeq();
			if ( !read && !read_mate )
			{
				return false;
			}
			else if ( read && read_mate )
			{
				if ( GetReadName( __fwd.GetSeqId()) ==   GetReadName( __bwd.GetSeqId() ) )
				{
					
					_SetSeqId     (__fwd.GetSeqId());
					_SetBase     ((unsigned char*)__fwd.GetBase());
					_SetSeqLength (__fwd.GetSeqLength());
					//Setting the fragment number (mandatory for a correct sam output format)
					__fwd.SetFragNum(FWD);
					__bwd.SetFragNum(BWD);		
					
					
					return true;
				}
				else
				{
					string query_id = __fwd.GetSeqId();
					string mate_id  = __bwd.GetSeqId();
					string msg  = "ERROR - query id = ";
					msg += query_id;
					msg += " but mate id = ";
					msg += mate_id;
					msg += '\n';
					throw (rtstack(msg));
				}
			}
			else
			{
				string msg = "ERROR - Not the same number of sequences in both files";
				throw (rtstack(msg));
			}
		};				

	    /* Several accessors */
		size_t GetSeqLengthBwd()            const { return __bwd.GetSeqLength(); };
	    const unsigned char * GetBaseBwd()  const { return __bwd.GetBase(); };
		auto_ptr<string> GetSeqAsFastaBwd() const { return __bwd.GetSeqAsFasta(); };
		unsigned char GetcBwd(uint32_t i)      const { return __bwd.Getc(i); };
	    uint64_t MaxScoreBwd(int matchscore)   const { return __bwd.MaxScore(matchscore); };
		
	private:
		T  __fwd;
		T  __bwd;
};

typedef QueryBufferFormatDouble<QueryBufferFasta> QueryBufferFastaDouble;
typedef QueryBufferFormatDouble<QueryBufferFastq> QueryBufferFastqDouble;

/**
 * \brief QueryBufferSamDouble: This QueryBuffer is useful for managing pair-ended sequences from a BAM (NOT SAM) file
 *                              where the reads and mates are in two different lines
 *                              In the constructor we read and index the file, then ReadNextSeq calls bam_seek to 
 *                              read both sequences (using two QueryBufferSam objects).
 *                             
 */
class QueryBufferSamDouble: public QueryBufferDouble
{
	public:
		QueryBufferSamDouble(const string&, bool qual=false, bool hsp=false);
		virtual QueryBufferSamDouble* Clone() { return new QueryBufferSamDouble( *this ); };
		~QueryBufferSamDouble(){};

		bool ReadNextSeq();
		
	    /* Several accessors */
		size_t GetSeqLengthBwd()            const { return __bwd.GetSeqLength(); };
	    const unsigned char * GetBaseBwd()  const { return __bwd.GetBase(); };
		auto_ptr<string> GetSeqAsFastaBwd() const { return __bwd.GetSeqAsFasta(); };
		unsigned char GetcBwd(uint32_t i)      const { return __bwd.Getc(i); };
	    uint64_t MaxScoreBwd(int matchscore)   const { return __bwd.MaxScore(matchscore); };

		virtual const Hsp* GetHsp()     const { return __fwd.GetHsp(); };
		virtual const Hsp* GetHspMate() const { return __bwd.GetHsp(); };
	    
	    QueryBufferSam& GetFwdQueryBuffer()  { return __fwd; };
	    QueryBufferSam& GetBwdQueryBuffer() { return __bwd; };
	    		
	    int GetOrphansNb() const { return __orphans.size(); };
	    
	private:
		QueryBufferBamLowLevel __low_level;
		QueryBufferSam __fwd;
		QueryBufferSam __bwd;
		static map<string,bam1_t*> __orphans;
		bool __hsp_flg;
};

/**
 * \brief QueryBufferBamLowLevelDouble: A LowLevel QueryBufferBamDouble, used by CommandSlice
 *        -ReadNextSeq()
 *        -GetLowLevelBufferFwd()
 *        -GetLowLevelBufferBwd()
 */

class QueryBufferBamLowLevelDouble: public QueryBufferDouble
{
	public:
		QueryBufferBamLowLevelDouble(const string& file_name);
		virtual QueryBufferBamLowLevelDouble* Clone() { return new QueryBufferBamLowLevelDouble( *this ); };
		~QueryBufferBamLowLevelDouble(){};
	    bam_header_t* GetHeader() const { return __low_level.GetHeader(); };

		virtual size_t GetSeqLengthBwd() const       {return 0; };
	    virtual const unsigned char * GetBaseBwd()  const { return NULL; };
		virtual auto_ptr<string> GetSeqAsFastaBwd() const { return auto_ptr<string>(new string(""));};
		virtual unsigned char GetcBwd(uint32_t i)      const { return '\0'; };
	    virtual uint64_t MaxScoreBwd(int matchscore)   const { return 0; };
	    virtual QueryBuffer& GetFwdQueryBuffer()    { return __low_level; };
	    virtual QueryBuffer& GetBwdQueryBuffer()    { return __low_level; };

		bool ReadNextSeq();
		
	    /* Several accessors */
	    const bam1_t* GetFwdLowLevelBuffer()  { return __fwd; };
	    const bam1_t* GetBwdLowLevelBuffer()  { return __bwd; };
	    
	private:
		bam1_t* __fwd;
		bam1_t* __bwd;
		QueryBufferBamLowLevel __low_level;
		static map<string,bam1_t*> __orphans;
};	


//bool  _SameRead( const string fwdId, const string bwdId );

#endif
