/*
   $Id: hsp.cpp 710 2013-01-29 12:34:31Z manu $
*/
/*
 * Copyright INRA/CNRS - Thomas.Faraut@toulouse.inra.fr, Emmanuel.Courcelle@toulouse.inra.fr
 * This software is a computer program whose purpose is to align nucleic sequences, or perform other utility tasks
 * This software is governed by the CeCILL-C license under French law and abiding by the rules of distribution of free software.
 * 
 * Please read the file Licence_CeCILL-C_V1-en.txt which should have been distributed with this file
 * The fact that you are presently reading this, and the aforementioned file, means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
*/


#include <ctype.h>
#include <iostream>
#include <list>
#include "utils.h"

#include "Parameters.h"
#include "hsp.h"
#include "GlintEngine.h"
#include "Display.h"

using namespace std;

extern RealLogger* logger;


#ifdef DEBUGLOG
void Hsp::LogHsp( const char* log_type, const char* header)
{
	logger -> SetLogType(log_type) << header << max_diag << '\t' << q_start << '\t' << q_end << '\t' << s_start << '\t' << s_end << '\t' << (int)__strand << '\t' << __score << '\n';
}
#endif

GapPile GapPile::Reverse(uint32_t hsp_len) const
{
	GapPile r_gaps;
	__Sort();
	for ( vector<Gap>::reverse_iterator i=gaps.rbegin(); i!=gaps.rend(); ++i )
	{
		r_gaps.Push(hsp_len - *i - 1);
	}
	return r_gaps;
}

/**
 * \brief We allocate HSP_PILE_SIZE hsps when building an HspPile
 * 
 * \param d The display object
*/

HspPile::HspPile(const Display& d) : 
	__parameters(Parameters::Instance()),
	top(0),
	hsp_prev_empty(true),
	display(d)
{
    hsp_prev    = new vector<Hsp*>;
    hsp_current = new vector<Hsp*>;
    hsp_prev    -> reserve ( HSP_PILE_SIZE );
    hsp_current -> reserve ( HSP_PILE_SIZE );
    for(int j=0; j<HSP_PILE_SIZE; j++) {
        hsp_current->push_back(new Hsp);
        hsp_prev->push_back(new Hsp);
    };
}

HspPile::~HspPile()
{
	while ( ! hsp_current->empty())
	{
		delete hsp_current->back();
		hsp_current->pop_back();
	}
	while ( ! hsp_prev->empty())
	{
		delete hsp_prev->back();
		hsp_prev->pop_back();
	}
	delete hsp_current;
	delete hsp_prev;
}		

/**
 * \brief Commit:
 *        - inserts the current hsp (ie the top) in l
 *        - increments top, WITHOUT increasing the size of the pile (as the hsps are preallocated)
 *        - If top is at the highest, the pile is dumped, we switch the piles and hsp_list is cleared
 */        

void HspPile::Commit()
{
	_Insert((*hsp_current)[top]);             // insert the top in l
	if (top < hsp_current->size()-1)
	{
        top++;
    }
    else
    {
		if (!__parameters.IsHspOnTheFly())
		{
			__parameters.Warn ("hsps are being displayed because the hsp pile is full, they are probably NOT correctly sorted");
#ifdef DEBUGLOG
			logger -> SetLogType ( "H" ) << " ** Warning: hsps are being displayed because the hsp pile is full.\n";
#endif
	        display.DisplayHits(*hsp_current);
		}
        hsp_list.clear();
        SwitchPiles();
    };
}

void HspPile::SwitchPiles()
{
	vector<Hsp*>* hv = hsp_prev;

	cerr<<"Hsp pile full => here switching pile \n";
	
	//We first sort the current hsp pile according to query start and then switch piles
	SortQuery();
	
	hsp_prev = hsp_current;
	hsp_current = hv;
	hsp_prev_empty = false;
	top=0;

}

/*---------------------------------------------*/
/**            The Lookup routines            **/
/*---------------------------------------------*/

/* The Lookup routines test if the corresponding match or hsp falls into an already   */
/* selected region as defined by the current and previous hsp list                    */
/* The already detected region are defined by min_diag, max_diag, s_start, s_end of a */
/* registered hsp, hence a parallelogram in the dotplot matrix                        */

/*
 Name:      Lookup
 Usage:     if (Lookup(m))
 Function:  returns true if both _LookupCurrent and _LookupPrevious are true
 Arg:       a pointer on a match
 Return:    a boolean
*/

bool HspPile::Lookup (const match& m) {
    return ( _LookupCurrent(m.diag,m.spos) || _LookupPrevious(m.qpos,m.spos) );

	/*bool b_current = _LookupCurrent(m.diag,m.spos);
	bool b_previous = _LookupPrevious(m.qpos,m.spos);
	bool b = ( b_current|| b_previous );

	if ( b )
	{
		cout << "Match: A similar hsp has already been identified. current :"<<b_current<<" previous :"<<b_previous<<"\n";
		cout << m ;
		cout << "\n";
	}

	return b;*/
}

/*
 Name:      Lookup
 Usage:     if (Lookup(h))
 Function:  returns true if  _LookupCurrent or _LookupPrevious is true
 Arg:       a pointer on a hsp
 Return:    a boolean
*/

bool HspPile::Lookup (const Hsp& h) {
	return (_LookupCurrent(h.max_diag,h.s_end) || _LookupPrevious(h.q_start,h.s_start));

	/*bool b_current = _LookupCurrent(h.max_diag,h.s_end);
	bool b_previous = _LookupPrevious(h.q_start,h.s_start);
	bool b = ( b_current|| b_previous );

	if (b)
	{
		cout << "Hsp: A similar hsp has already been identified. current :"<<b_current<<" previous :"<<b_previous<<"\n";
		cout << h << "\n";
	}

	return b;*/
}


/*
 Name:		_LookupCurrent
 Usage:     _LookupCurrent(diag,spos)
 Function:  Returns true if (diag,pos) is contained in the parallelogram (sstart,min_diag) <-> (send,max_diag) of one of the hsps of a list.
            The list of hsp searched is the current list hsp_list (sorted by min_diag)
 Arg:       a diagonal and a position on the subject
 Return:    a boolean
 TODO:      Should be modified in order to check for both s_start and s_end when an hsp is tested !!!!
*/

bool HspPile::_LookupCurrent(int32_t diag, uint32_t spos)
{
	list<Hsp*>::iterator i=hsp_list.begin();
    while (i!=hsp_list.end() && (diag<(**i).max_diag || ( diag==(**i).max_diag && spos<=(**i).s_end )))
    {    															 // while there is a potential overlap
        ++i;
    };

	list<Hsp*>::reverse_iterator j(i); // Warning j points now to the hsp preceding i, because i is out of bounds :
	                                   // diag>(**i).max_diag or diag==(**i).max_diag && spos > (**i).s_end

	while (j!=hsp_list.rend() &&  diag<=(**j).max_diag)
    {  																// while there is overlap, considering only the diagonals
    	if (diag>=(**j).min_diag && spos>=(**j).s_start && spos<=(**j).s_end)
        {
		    return true;     										// Overlap detected !!!
        };
        ++j;
    };

    return false;
}

/*
 Name:		_LookupPrevious
 Usage:     _LookupPrevious(qpos,spos)
 Function:  Returns true is (qpos,pos) is contained in the parallelogram (qstart,sstart) <-> (qend,send) of one of the hsps of a list.
            The list of hsp searched is the previous list hsp_prev. In this list, hsps are sorted according to qstart (...).
 Arg:       a psoition on the query and a position on the subject
 Return:    a boolean
 TODO:      Should be modified in order to check for both s_start and s_end when an hsp is tested !!!!
*/

bool HspPile::_LookupPrevious(uint32_t qpos, uint32_t spos)
{
	Hsp *h=(*hsp_prev)[0];

	/*
    if (Debug_Here) {
        fprintf(stdout,"%d %d %d  in %d  %d  %d  %d  ? :  %d\n",hsp_prev->size()i,qpos,spos,h->q_start,h->q_end,h->s_start,h->s_end,1);
		for (int i=0;i<hsp_prev->size();i++) {
			display_hsp_tabular(hsp_prev[i],stdout);
		}
        fprintf(stdout,"----------------\n");
    };
    */

	if (hsp_prev_empty || qpos > h->q_end)
        return false;

    uint32_t i=1;
	while ( i < hsp_prev->size() && qpos < h->q_end ) {
		if ( qpos >= h->q_start ) {
			if (spos >= h->s_start && spos <= h->s_end) {
                //fprintf(stdout,"poverlap: %d %d  in %d  %d  %d  %d\n",qpos,spos,h->q_start,h->q_end,h->s_start,h->s_end);
                return true;
            }
		}
		h=(*hsp_prev)[i++];
	}
    return false;
}


#ifdef DEBUGLOG
void HspPile::LogPile ( const char* log_type )
{
	for (list<Hsp*>::iterator i=hsp_list.begin(); i != hsp_list.end(); ++i)
    {
		(*i) -> LogHsp(log_type, "hsp pile ");
	}
}
#endif

/*
     operator< for the hsp, used for sorting the hsps
        1st key: Name (increasing order)
        2nd key: score (decreasing order)

*/
bool operator<(const Hsp& p, const Hsp& q)
{
    if (p.Subject.Name == q.Subject.Name)
        return (p.GetScore() > q.GetScore());
    else
        return (p.Subject.Name < q.Subject.Name);
}


/* A very basic ostream operator for hsp */

/*ostream & operator<<(ostream& os, const Hsp& h)
{
	const char *char_sign=(h.strand >0) ? OUTPUT_PLUS:OUTPUT_MINUS;
	os<<h.min_diag<<"\t"<<h.max_diag<<"\t"<<h.q_start<<"\t"<<h.q_end<<"\t"<<h.s_start<<"\t"<<h.s_end<<"\t"<<char_sign<<"\t"<<h.score<<"\t"<<h.size;
	return os;
}*/

ostream & operator<<(ostream& os, const Hsp& h)
{
	const char *char_sign=(h.__strand >0) ? OUTPUT_PLUS:OUTPUT_MINUS;
	os<<h.Query.Name<<"\t"<<h.q_start<<"\t"<<h.q_end<<"\t"<<h.Subject.Name<<"\t"<<h.s_start<<"\t"<<h.s_end<<"\t"<<char_sign<<"\t"<<h.__score<<"\t"<<h.__size;
	return os;
}

/**
 * \brief HitPile constructor
 * \param v A vector of Hsp* - As a side effect, this vector will be SORTED !!!
 *         best_score A boolean, if true only the best_score hsps are kept for each hit
 */

HitPile::HitPile ( vector<Hsp*>& v, bool best_score)
{
	if ( v.empty() ) return;
	if ( best_score)
	{
		// Sorting the hps on score
		sort(v.begin(),v.end(),CompHspScorePtr);
		uint32_t bestScore = v[0]->GetScore();
		
		for (size_t i=0; i<v.size() && v[i]->GetScore() == bestScore; i++)
		{	
			__AddSortedHspPtr(v[i]);		
		}
		//Don't need to sort the hits here, all the hits have the same score
	}
	else
	{
		// Sorting the hps, 1st key name, 2nd key score
		sort(v.begin(),v.end(),CompHspPtr);

		for (size_t i=0; i<v.size(); i++)
		{
			__AddSortedHspPtr(v[i]);
		}
		__hits.sort();
	}
}

/**
 * \brief Add the hsp_ptr to the last Hit, may be creating a new last Hit
 *        If the hsp_ptr as a Name different from the last Hit name, a new hit is created
 * \pre   The hsp must be sorted before continuously adding them
 * \param hsp_ptr a new hsp_ptr to add
 */

void HitPile::__AddSortedHspPtr(Hsp* hsp_ptr)
{
	if ( __hits.empty() || hsp_ptr->Subject.Name != __hits.back().Name )
	{
		__NewHit( hsp_ptr );
	}
	__hits.back().hsps.push_back(hsp_ptr);
}

